"""Module to deal with orbit acquisition."""
import os as _os
import time as _time
from math import ceil as _ceil
import logging as _log
from functools import partial as _part
from copy import deepcopy as _dcopy
from threading import Lock, Thread, Event as _Event
import multiprocessing as _mp
import traceback as _traceback

import numpy as _np
import bottleneck as _bn

from .. import util as _util
from ..diagbeam.bpm.csdev import Const as _csbpm
from ..thread import RepeaterThread as _Repeat
from ..epics import PV as _PV, CAProcessSpawn as _Process

from .base_class import BaseClass as _BaseClass
from .bpms import BPM, TimingConfig, TIMEOUT


class BaseOrbit(_BaseClass):
    """."""


def run_subprocess(pvs, send_pipe, recv_pipe):
    """Run subprocesses."""
    max_spread = 25/1000  # in [s]
    timeout = 50/1000  # in [s]

    ready_evt = _Event()

    tstamps = _np.full(len(pvs), _np.nan)

    def callback(*_, **kwargs):
        pvo = kwargs['cb_info'][1]
        # pvo._args['timestamp'] = _time.time()
        tstamps[pvo.index] = pvo.timestamp
        maxi = _bn.nanmax(tstamps)
        mini = _bn.nanmin(tstamps)
        if (maxi-mini) < max_spread:
            ready_evt.set()

    def conn_callback(pvname=None, conn=None, pv=None):
        if not conn:
            tstamps[pv.index] = _np.nan

    pvsobj = []
    for i, pvn in enumerate(pvs):
        pvo = _PV(pvn, connection_timeout=TIMEOUT)
        pvo.index = i
        pvsobj.append(pvo)

    for pvo in pvsobj:
        pvo.wait_for_connection()

    for pvo in pvsobj:
        pvo.add_callback(callback)
        pvo.connection_callbacks.append(conn_callback)

    boo = True
    while boo or recv_pipe.recv():
        boo = False
        ready_evt.clear()
        nok = 0.0
        if not ready_evt.wait(timeout=timeout):
            nok = 1.0
        out = []
        for pvo in pvsobj:
            if not pvo.connected:
                out.append(_np.nan)
                continue
            # out.append(pvo.timestamp)
            out.append(pvo.value)
        out.append(nok)
        send_pipe.send(out)


class EpicsOrbit(BaseOrbit):
    """Class to deal with orbit acquisition."""

    def __init__(self, acc, prefix='', callback=None):
        """Initialize the instance."""
        super().__init__(acc, prefix=prefix, callback=callback)

        self._mode = 0  # first mode of the list
        self._sofb = None
        self._sync_with_inj = False
        self._sloworb_timeout = 0
        self.ref_orbs = {
            'X': _np.zeros(self._csorb.nr_bpms),
            'Y': _np.zeros(self._csorb.nr_bpms)}
        self._load_ref_orbs()
        self.raw_orbs = {'X': [], 'Y': []}
        self.raw_sporbs = {'X': [], 'Y': [], 'Sum': []}
        self.raw_mtorbs = {'X': [], 'Y': [], 'Sum': []}
        self._lock_raw_orbs = Lock()
        self.smooth_orb = {'X': None, 'Y': None}
        self.smooth_sporb = {'X': None, 'Y': None, 'Sum': None}
        self.smooth_mtorb = {'X': None, 'Y': None, 'Sum': None}
        self._smooth_npts = 1
        self._smooth_meth = self._csorb.SmoothMeth.Average
        self._spass_mask = [0, 0]
        self._spass_average = 1
        self._acqtrignrsamplespre = 0
        self._acqtrignrsamplespost = 360
        self._acqtrignrshots = 1
        self._multiturnidx = 0
        self._mturndownsample = 1
        self._timevector = None
        self.bpms = [BPM(name, callback) for name in self._csorb.bpm_names]
        self.timing = TimingConfig(acc, callback)
        self.new_orbit = _Event()
        if self.acc == 'SI':
            self._processes = []
            self._mypipes_recv = []
            self._mypipes_send = []
            self._create_processes(nrprocs=16)
        self._orbit_thread = _Repeat(
            1/self._csorb.ACQRATE_SLOWORB, self._update_orbits, niter=0)
        self._orbit_thread.start()
        self._thread_sync = None
        self._update_time_vector()

    @property
    def sofb(self):
        """."""
        return self._sofb

    @sofb.setter
    def sofb(self, sofb):
        self._sofb = sofb

    def _create_processes(self, nrprocs=8):
        # get the start method of the Processes that will be launched:
        spw = _mp.get_context('spawn')

        pvs = []
        for bpm in self._csorb.bpm_names:
            pvs.append(bpm+':PosX-Mon')
        for bpm in self._csorb.bpm_names:
            pvs.append(bpm+':PosY-Mon')

        # subdivide the pv list for the processes
        div = len(pvs) // nrprocs
        rem = len(pvs) % nrprocs
        sub = [div*i + min(i, rem) for i in range(nrprocs+1)]

        # create processes
        for i in range(nrprocs):
            mine, send_pipe = spw.Pipe(duplex=False)
            self._mypipes_recv.append(mine)
            recv_pipe, mine = spw.Pipe(duplex=False)
            self._mypipes_send.append(mine)
            pvsn = pvs[sub[i]:sub[i+1]]
            self._processes.append(_Process(
                target=run_subprocess,
                args=(pvsn, send_pipe, recv_pipe),
                daemon=True))
        for proc in self._processes:
            proc.start()

    def shutdown(self):
        """."""
        self._orbit_thread.resume()
        self._orbit_thread.stop()
        self._orbit_thread.join()
        if self.acc == 'SI':
            for pipe in self._mypipes_send:
                pipe.send(False)
                pipe.close()
            for proc in self._processes:
                proc.join()

    def get_map2write(self):
        """Get the write methods of the class."""
        dbase = {
            'SOFBMode-Sel': self.set_orbit_mode,
            'SyncWithInjection-Sel': self.set_sync_with_injection,
            'TrigAcqConfig-Cmd': self.acq_config_bpms,
            'TrigAcqCtrl-Sel': self.set_trig_acq_control,
            'TrigAcqChan-Sel': self.set_trig_acq_channel,
            'TrigAcqRepeat-Sel': self.set_trig_acq_repeat,
            'TrigNrSamplesPre-SP': _part(self.set_acq_nrsamples, ispost=False),
            'TrigNrSamplesPost-SP': _part(self.set_acq_nrsamples, ispost=True),
            'RefOrbX-SP': _part(self.set_reforb, 'X'),
            'RefOrbY-SP': _part(self.set_reforb, 'Y'),
            'SmoothNrPts-SP': self.set_smooth_npts,
            'SmoothMethod-Sel': self.set_smooth_method,
            'SmoothReset-Cmd': self.set_smooth_reset,
            'SPassMaskSplBeg-SP': _part(self.set_spass_mask, beg=True),
            'SPassMaskSplEnd-SP': _part(self.set_spass_mask, beg=False),
            'SPassAvgNrTurns-SP': self.set_spass_average,
            'TrigNrShots-SP': self.set_trig_acq_nrshots,
            'PolyCalibration-Sel': self.set_poly_calibration,
            'SyncBPMs-Cmd': self.sync_bpms,
            }
        if not self.isring:
            return dbase
        dbase.update({
            'MTurnAcquire-Cmd': self.acquire_mturn_orbit,
            'MTurnIdx-SP': self.set_orbit_multiturn_idx,
            'MTurnDownSample-SP': self.set_mturndownsample,
            'MTurnSyncTim-Sel': self.set_mturn_sync,
            'MTurnUseMask-Sel': self.set_mturn_usemask,
            'MTurnMaskSplBeg-SP': _part(self.set_mturnmask, beg=True),
            'MTurnMaskSplEnd-SP': _part(self.set_mturnmask, beg=False),
            })
        return dbase

    @property
    def mode(self):
        """."""
        return self._mode

    @property
    def acqtrignrsamples(self):
        """."""
        return self._acqtrignrsamplespre + self._acqtrignrsamplespost

    @property
    def update_raws(self):
        """."""
        return not self._sync_with_inj or self.timing.injecting

    def is_sloworb(self, mode=None):
        """Check is mode or self._mode is in SlowOrb mode."""
        if mode is None:
            mode = self._mode
        return self.acc == 'SI' and mode == self._csorb.SOFBMode.SlowOrb

    def is_multiturn(self, mode=None):
        """Check is mode or self._mode is in MultiTurn mode."""
        if mode is None:
            mode = self._mode
        return self.isring and mode == self._csorb.SOFBMode.MultiTurn

    def is_singlepass(self, mode=None):
        """Check is mode or self._mode is in SinglePass mode."""
        if mode is None:
            mode = self._mode
        return mode == self._csorb.SOFBMode.SinglePass

    def is_trigmode(self, mode=None):
        """Check is mode or self._mode is in any of the Triggered modes."""
        return self.is_singlepass(mode) or self.is_multiturn(mode)

    def get_orbit(self, reset=False, synced=False, timeout=1/10):
        """Return the orbit distortion."""
        nrb = self._csorb.nr_bpms
        refx = self.ref_orbs['X'][:nrb]
        refy = self.ref_orbs['Y'][:nrb]

        if reset:
            with self._lock_raw_orbs:
                self._reset_orbs()
            _time.sleep(self._smooth_npts/self._csorb.ACQRATE_SLOWORB)

        if self.is_multiturn():
            orbs = self.smooth_mtorb
            raws = self.raw_mtorbs
            getorb = self._get_orbit_multiturn
        elif self.is_singlepass():
            orbs = self.smooth_sporb
            raws = self.raw_sporbs
            getorb = self._get_orbit_singlepass
        elif self.is_sloworb():
            if synced:
                self.new_orbit.wait(timeout=timeout)
                self.new_orbit.clear()
            orbs = self.smooth_orb
            raws = self.raw_orbs
            getorb = self._get_orbit_online

        for _ in range(3 * self._smooth_npts):
            with self._lock_raw_orbs:
                isempty = orbs['X'] is None or orbs['Y'] is None
                if not isempty and len(raws['X']) >= self._smooth_npts:
                    orbx, orby = getorb(orbs)
                    break
            _time.sleep(1/self._csorb.ACQRATE_SLOWORB)
        else:
            msg = 'ERR: timeout waiting orbit.'
            self._update_log(msg)
            _log.error(msg[5:])
            orbx, orby = refx, refy
        # # for tests:
        # orbx -= _time.time()
        # orby -= _time.time()
        return _np.hstack([orbx-refx, orby-refy])

    def _get_orbit_online(self, orbs):
        """."""
        return orbs['X'], orbs['Y']

    def _get_orbit_singlepass(self, orbs):
        """."""
        return orbs['X'], orbs['Y']

    def _get_orbit_multiturn(self, orbs):
        """."""
        idx = self._multiturnidx
        return orbs['X'][idx, :], orbs['Y'][idx, :]

    def set_smooth_npts(self, num):
        """."""
        self._smooth_npts = num
        self.run_callbacks('SmoothNrPts-RB', num)
        return True

    def set_smooth_method(self, meth):
        """."""
        self._smooth_meth = meth
        self.run_callbacks('SmoothMethod-Sts', meth)
        return True

    def set_spass_mask(self, val, beg=True):
        """."""
        val = int(val) if val > 0 else 0
        other_mask = self._spass_mask[1 if beg else 0]
        maxsz = self.bpms[0].tbtrate - other_mask - 2
        val = val if val < maxsz else maxsz
        self._spass_mask[0 if beg else 1] = val
        name = 'Beg' if beg else 'End'
        self.run_callbacks('SPassMaskSpl' + name + '-RB', val)
        return True

    def set_mturn_sync(self, val):
        """."""
        value = _csbpm.EnbldDsbld.enabled
        if val == self._csorb.EnbldDsbld.Dsbld:
            value = _csbpm.EnbldDsbld.disabled

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.tbt_sync_enbl = value
        self.run_callbacks('MTurnSyncTim-Sts', val)
        return True

    def set_mturn_usemask(self, val):
        """."""
        value = _csbpm.EnbldDsbld.enabled
        if val == self._csorb.EnbldDsbld.Dsbld:
            value = _csbpm.EnbldDsbld.disabled

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.tbt_mask_enbl = value

        self.run_callbacks('MTurnUseMask-Sts', val)
        return True

    def set_mturnmask(self, val, beg=True):
        """."""
        val = int(val) if val > 0 else 0
        bpms = self._get_used_bpms()
        omsk = \
            bpms[0].tbt_mask_begin if not beg else \
            bpms[0].tbt_mask_end
        omsk = omsk or 0
        maxsz = bpms[0].tbtrate - omsk - 2
        val = val if val < maxsz else maxsz

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            if beg:
                bpm.tbt_mask_begin = val
            else:
                bpm.tbt_mask_end = val

        name = 'Beg' if beg else 'End'
        self.run_callbacks('MTurnMaskSpl' + name + '-RB', val)
        return True

    def set_spass_average(self, val):
        """."""
        val = int(val) if val > 1 else 1
        with self._lock_raw_orbs:
            self._spass_average = val
            self._reset_orbs()
        self.run_callbacks('SPassAvgNrTurns-RB', val)
        Thread(target=self._prepare_mode, daemon=True).start()
        return True

    def set_smooth_reset(self, _):
        """."""
        with self._lock_raw_orbs:
            self._reset_orbs()
        return True

    def set_reforb(self, plane, orb):
        """."""
        msg = 'Setting New Reference Orbit.'
        self._update_log(msg)
        _log.info(msg)
        orb = _np.array(orb, dtype=float)
        nrb = self._csorb.nr_bpms
        if orb.size % self._csorb.nr_bpms:
            msg = 'ERR: Wrong RefOrb Size.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        elif orb.size < nrb:
            msg = 'WARN: Orb Size is too small. Replicating...'
            self._update_log(msg)
            _log.error(msg[6:])
            nrep = int(nrb//orb.size) + 1
            orb2 = _np.tile(orb, nrep)
            orb = orb2[:nrb]
        self.ref_orbs[plane] = orb
        self._save_ref_orbits()
        with self._lock_raw_orbs:
            self._reset_orbs()
        self.run_callbacks('RefOrb'+plane+'-RB', orb[:nrb])
        return True

    def set_orbit_mode(self, value):
        """."""
        omode = self._mode
        acqrate = self._csorb.ACQRATE_SLOWORB
        if self.is_trigmode(value):
            acqrate = self._csorb.ACQRATE_TRIGMODE

        with self._lock_raw_orbs:
            self._mode = value
            self._orbit_thread.interval = 1/acqrate
            self._reset_orbs()
        Thread(
            target=self._prepare_mode,
            kwargs={'oldmode': omode, }, daemon=True).start()
        self.run_callbacks('SOFBMode-Sts', value)
        return True

    def set_sync_with_injection(self, boo):
        """."""
        self._sync_with_inj = bool(boo)
        self.run_callbacks('SyncWithInjection-Sts', bool(boo))
        return True

    def _prepare_mode(self, oldmode=None):
        """."""
        oldmode = self._mode if oldmode is None else oldmode
        self.set_trig_acq_control(self._csorb.TrigAcqCtrl.Abort)

        if not self.is_trigmode():
            self.acq_config_bpms()
            return True

        if self.is_singlepass():
            chan = self._csorb.TrigAcqChan.ADCSwp
            rep = self._csorb.TrigAcqRepeat.Repetitive
            points = self._spass_average * self.bpms[0].tbtrate
        elif self.is_multiturn():
            chan = self._csorb.TrigAcqChan.TbT
            rep = self._csorb.TrigAcqRepeat.Repetitive
            points = self._mturndownsample

        if self._mode != oldmode:
            self.run_callbacks('TrigAcqChan-Sel', chan)
            self.set_trig_acq_channel(chan)
            self.run_callbacks('TrigAcqRepeat-Sel', rep)
            self.set_trig_acq_repeat(rep)
            if self.acqtrignrsamples < points:
                pts = points - self._acqtrignrsamplespre
                self.run_callbacks('TrigNrSamplesPost-SP', pts)
                self.set_acq_nrsamples(pts, ispost=True)
        self._update_time_vector()
        self.acq_config_bpms()

        self.set_trig_acq_control(self._csorb.TrigAcqCtrl.Start)
        return True

    def set_orbit_multiturn_idx(self, value):
        """."""
        maxidx = self.acqtrignrsamples // self._mturndownsample
        maxidx *= self._acqtrignrshots
        if value >= maxidx:
            value = maxidx-1
            msg = 'WARN: MTurnIdx is too large. Redefining...'
            self._update_log(msg)
            _log.warning(msg[6:])
        with self._lock_raw_orbs:
            self._multiturnidx = int(value)
        self.run_callbacks('MTurnIdx-RB', self._multiturnidx)
        self.run_callbacks(
            'MTurnIdxTime-Mon', self._timevector[self._multiturnidx])
        return True

    def acq_config_bpms(self, *args):
        """."""
        _ = args
        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            if self.is_multiturn():
                bpm.mode = _csbpm.OpModes.MultiBunch
                bpm.switching_mode = _csbpm.SwModes.direct
                bpm.configure()
                self.timing.configure()
            elif self.is_singlepass():
                bpm.mode = _csbpm.OpModes.MultiBunch
                bpm.switching_mode = _csbpm.SwModes.direct
                bpm.configure()
                self.timing.configure()
            elif self.is_sloworb():
                bpm.switching_mode = _csbpm.SwModes.switching
        self.sync_bpms(None)
        return True

    def sync_bpms(self, *args):
        """Synchronize BPMs."""
        _ = args

        msg = 'Received sync BPMs command...'
        self._update_log(msg)
        _log.info(msg)

        if self._thread_sync is not None and \
                self._thread_sync.is_alive():
            msg = 'Waiting for previous sync to end...'
            self._update_log(msg)
            _log.info(msg)
            self._thread_sync.join()

        self._thread_sync = Thread(
            target=self._synchronize_bpms, daemon=True)
        self._thread_sync.start()

    def _synchronize_bpms(self):
        msg = 'Syncing BPMs...'
        self._update_log(msg)
        _log.info(msg)
        for bpm in self._get_used_bpms():
            bpm.tbt_sync_enbl = _csbpm.EnbldDsbld.enabled
            bpm.fofb_sync_enbl = _csbpm.EnbldDsbld.enabled
            bpm.monit1_sync_enbl = _csbpm.EnbldDsbld.enabled
            bpm.monit_sync_enbl = _csbpm.EnbldDsbld.enabled
        _time.sleep(0.5)
        for bpm in self._get_used_bpms():
            bpm.tbt_sync_enbl = _csbpm.EnbldDsbld.disabled
            bpm.fofb_sync_enbl = _csbpm.EnbldDsbld.disabled
            bpm.monit_sync_enbl = _csbpm.EnbldDsbld.disabled
            bpm.monit1_sync_enbl = _csbpm.EnbldDsbld.disabled

        if self.acc == 'SI' and self.sofb.fofb.connected:
            _time.sleep(0.2)
            msg = 'Syncing FOFB Net...'
            self._update_log(msg)
            _log.info(msg)
            self.sofb.fofb.cmd_fofbctrl_syncnet()

        msg = '...done!'
        self._update_log(msg)
        _log.info(msg)

    def set_trig_acq_control(self, value):
        """."""
        acqctrl = self._csorb.TrigAcqCtrl
        if value == acqctrl.Start:
            self._wait_bpms()

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.ctrl = value
        self.run_callbacks('TrigAcqCtrl-Sts', value)

        if value in {acqctrl.Stop, acqctrl.Abort}:
            self._wait_bpms()
        return True

    def set_trig_acq_channel(self, value):
        """."""
        try:
            val = self._csorb.TrigAcqChan._fields[value]
            val = _csbpm.AcqChan._fields.index(val)
        except (IndexError, ValueError):
            return False

        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.acq_type = val

        self.run_callbacks('TrigAcqChan-Sts', value)
        self._update_time_vector(channel=val)
        return True

    def set_trig_acq_repeat(self, value):
        """."""
        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.acq_repeat = value

        self.run_callbacks('TrigAcqRepeat-Sts', value)
        return True

    def set_acq_nrsamples(self, val, ispost=True):
        """."""
        val = int(val) if val > 0 else 0
        val = val if val < 20000 else 20000
        suf = 'post' if ispost else 'pre'
        oth = 'post' if not ispost else 'pre'
        if getattr(self, '_acqtrignrsamples' + oth) == 0 and val == 0:
            return False
        with self._lock_raw_orbs:
            mask = self._get_mask()
            for i, bpm in enumerate(self.bpms):
                bpm.put_enable = mask[i]
                setattr(bpm, 'nrsamples' + suf, val)
            self._reset_orbs()
            setattr(self, '_acqtrignrsamples' + suf, val)
        self.run_callbacks('TrigNrSamples'+suf.title()+'-RB', val)
        self._update_time_vector()
        return True

    def set_trig_acq_nrshots(self, val):
        """."""
        val = int(val) if val > 1 else 1
        val = val if val < 1000 else 1000
        with self._lock_raw_orbs:
            mask = self._get_mask()
            for i, bpm in enumerate(self.bpms):
                bpm.put_enable = mask[i]
                bpm.nrshots = val
            self.timing.nrpulses = val
            self._reset_orbs()
            self._acqtrignrshots = val
        self.run_callbacks('TrigNrShots-RB', val)
        self._update_time_vector()
        return True

    def set_poly_calibration(self, val):
        """."""
        value = _csbpm.EnbldDsbld.enabled
        if val == self._csorb.EnbldDsbld.Dsbld:
            value = _csbpm.EnbldDsbld.disabled
        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            bpm.polycal = value
        self.run_callbacks('PolyCalibration-Sts', val)
        return True

    def set_mturndownsample(self, val):
        """."""
        val = int(val) if val > 1 else 1
        val = val if val < 1000 else 1000
        with self._lock_raw_orbs:
            self._mturndownsample = val
            self._reset_orbs()
        self.run_callbacks('MTurnDownSample-RB', val)
        Thread(target=self._prepare_mode, daemon=True).start()
        return True

    def acquire_mturn_orbit(self, _):
        """Acquire Multiturn data from BPMs."""
        Thread(target=self._update_multiturn_orbits, daemon=True).start()

    def _wait_bpms(self):
        """."""
        for _ in range(40):
            if all(map(lambda x: x.is_ok, self._get_used_bpms())):
                break
            _time.sleep(0.1)
        else:
            _log.warning('Timeout waiting BPMs.')

    def _update_time_vector(self, delay=None, duration=None, channel=None):
        """."""
        if not self.isring:
            return
        dly = (delay or self.timing.totaldelay or 0.0) * 1e-6  # from us to s
        dur = (duration or self.timing.duration or 0.0) * 1e-6  # from us to s
        channel = channel or self.bpms[0].acq_type or 0

        # revolution period in s
        if channel == _csbpm.AcqChan.Monit1:
            dtime = self.bpms[0].monit1period
        elif channel == _csbpm.AcqChan.FOFB:
            dtime = self.bpms[0].fofbperiod
        else:
            dtime = self.bpms[0].tbtperiod
        mult = self._mturndownsample
        dtime *= mult
        nrptpst = self.acqtrignrsamples // mult
        offset = self._acqtrignrsamplespre / mult
        nrst = self._acqtrignrshots
        shots = _np.arange(nrst)
        pts = _np.arange(nrptpst, dtype=float) + (0.5 - offset)
        vect = dly + dur/nrst*shots[:, None] + dtime*pts[None, :]
        self._timevector = vect.ravel()
        self.run_callbacks('MTurnTime-Mon', self._timevector)
        self.set_orbit_multiturn_idx(self._multiturnidx)

    def _load_ref_orbs(self):
        """."""
        if not _os.path.isfile(self._csorb.ref_orb_fname):
            return
        self.ref_orbs['X'], self.ref_orbs['Y'] = _np.loadtxt(
            self._csorb.ref_orb_fname, unpack=True)
        self.run_callbacks('RefOrbX-RB', self.ref_orbs['X'].copy())
        self.run_callbacks('RefOrbY-RB', self.ref_orbs['Y'].copy())

    def _save_ref_orbits(self):
        """."""
        refx = self.ref_orbs['X']
        refy = self.ref_orbs['Y']
        if refx.size < refy.size:
            ref = _np.zeros(refy.shape, dtype=float)
            ref[:refx.size] = refx
            refx = ref
        elif refy.size < refx.size:
            ref = _np.zeros(refx.shape, dtype=float)
            ref[:refy.size] = refy
            refy = ref
        orbs = _np.array([refx, refy]).T
        try:
            path = _os.path.split(self._csorb.ref_orb_fname)[0]
            _os.makedirs(path, exist_ok=True)
            _np.savetxt(self._csorb.ref_orb_fname, orbs)
        except FileNotFoundError:
            msg = 'WARN: Could not save reference orbit in file.'
            self._update_log(msg)
            _log.warning(msg[6:])

    def _reset_orbs(self):
        """."""
        raw = self.raw_orbs
        smt = self.smooth_orb
        raw['X'], raw['Y'] = [], []
        smt['X'], smt['Y'] = None, None
        raw = self.raw_sporbs
        smt = self.smooth_sporb
        raw['X'], raw['Y'], raw['Sum'] = [], [], []
        smt['X'], smt['Y'], smt['Sum'] = None, None, None
        raw = self.raw_mtorbs
        smt = self.smooth_mtorb
        raw['X'], raw['Y'], raw['Sum'] = [], [], []
        smt['X'], smt['Y'], smt['Sum'] = None, None, None
        self.run_callbacks('BufferCount-Mon', 0)

    def _update_orbits(self):
        """."""
        try:
            count = 0
            if self.is_multiturn():
                self._update_multiturn_orbits()
                count = len(self.raw_mtorbs['X'])
            elif self.is_singlepass():
                self._update_singlepass_orbits()
                count = len(self.raw_sporbs['X'])
            elif self.is_sloworb():
                self._update_online_orbits()
                count = len(self.raw_orbs['X'])
            self.run_callbacks('BufferCount-Mon', count)
        except Exception as err:
            self._update_log('ERR: ' + str(err))
            _log.error(_traceback.format_exc())

    def _update_online_orbits(self):
        """."""
        posx, posy, nok = self._get_orbit_from_processes()
        posx /= 1000
        posy /= 1000
        nanx = _np.isnan(posx)
        nany = _np.isnan(posy)
        posx[nanx] = self.ref_orbs['X'][nanx]
        posy[nany] = self.ref_orbs['Y'][nany]
        orbs = {'X': posx, 'Y': posy}

        for plane in ('X', 'Y'):
            with self._lock_raw_orbs:
                raws = self.raw_orbs
                raws[plane].append(orbs[plane])
                raws[plane] = raws[plane][-self._smooth_npts:]
                if not raws[plane]:
                    return
                if self._smooth_meth == self._csorb.SmoothMeth.Average:
                    orb = _np.mean(raws[plane], axis=0)
                else:
                    orb = _np.median(raws[plane], axis=0)
                self.smooth_orb[plane] = orb
        self.new_orbit.set()

        self._sloworb_timeout += nok
        if self._sloworb_timeout >= 1000:
            self._sloworb_timeout = 0
        self.run_callbacks('SlowOrbTimeout-Mon', self._sloworb_timeout)
        for plane in ('X', 'Y'):
            orb = self.smooth_orb[plane]
            if orb is None:
                return
            dorb = orb - self.ref_orbs[plane]
            self.run_callbacks(f'SlowOrb{plane:s}-Mon', _np.array(orb))
            self.run_callbacks(f'DeltaOrb{plane:s}Avg-Mon', _bn.nanmean(dorb))
            self.run_callbacks(f'DeltaOrb{plane:s}Std-Mon', _bn.nanstd(dorb))
            self.run_callbacks(f'DeltaOrb{plane:s}Min-Mon', _bn.nanmin(dorb))
            self.run_callbacks(f'DeltaOrb{plane:s}Max-Mon', _bn.nanmax(dorb))

    def _get_orbit_from_processes(self):
        nr_bpms = self._csorb.nr_bpms
        out = []
        nok = []
        for pipe in self._mypipes_recv:
            res = pipe.recv()
            out.extend(res[:-1])
            nok.append(res[-1])
        for pipe in self._mypipes_send:
            pipe.send(True)
        orbx = _np.array(out[:nr_bpms], dtype=float)
        orby = _np.array(out[nr_bpms:], dtype=float)
        return orbx, orby, any(nok)

    def _update_multiturn_orbits(self):
        """."""
        orbs = {'X': [], 'Y': [], 'Sum': []}
        with self._lock_raw_orbs:  # I need the lock here to ensure consistency
            samp = self.acqtrignrsamples
            down = self._mturndownsample
            samp -= samp % down
            if samp < 1:
                msg = 'ERR: Actual nr_samples in MTurn orb calc. is < 1.'
                self._update_log(msg)
                _log.error(msg[5:])
                return
            samp *= self._acqtrignrshots
            orbsz = self._csorb.nr_bpms
            nr_pts = self._smooth_npts
            for i, bpm in enumerate(self.bpms):
                pos = bpm.mtposx
                if pos is None:
                    posx = _np.full(samp, self.ref_orbs['X'][i])
                elif pos.size < samp:
                    posx = _np.full(samp, self.ref_orbs['X'][i])
                    posx[:pos.size] = pos
                else:
                    posx = pos[:samp]
                pos = bpm.mtposy
                if pos is None:
                    posy = _np.full(samp, self.ref_orbs['Y'][i])
                elif pos.size < samp:
                    posy = _np.full(samp, self.ref_orbs['Y'][i])
                    posy[:pos.size] = pos
                else:
                    posy = pos[:samp]
                pos = bpm.mtsum
                if pos is None:
                    psum = _np.full(samp, 0)
                elif pos.size < samp:
                    psum = _np.full(samp, 0)
                    psum[:pos.size] = pos
                else:
                    psum = pos[:samp]
                orbs['X'].append(posx)
                orbs['Y'].append(posy)
                orbs['Sum'].append(psum)

            for pln, raw in self.raw_mtorbs.items():
                norb = _np.array(orbs[pln], dtype=float)  # bpms x turns
                norb = norb.T.reshape(-1, orbsz)  # turns/rz x rz*bpms
                if self.update_raws:
                    raw.append(norb)
                del raw[:-nr_pts]
                if not raw:
                    return
                if self._smooth_meth == self._csorb.SmoothMeth.Average:
                    orb = _np.mean(raw, axis=0)
                else:
                    orb = _np.median(raw, axis=0)
                if down > 1:
                    orb = _np.mean(orb.reshape(-1, down, orbsz), axis=1)
                self.smooth_mtorb[pln] = orb
                orbs[pln] = orb
        idx = min(self._multiturnidx, orb.shape[0])
        for pln, orb in orbs.items():
            name = ('Orb' if pln != 'Sum' else '') + pln
            self.run_callbacks('MTurn' + name + '-Mon', orb.ravel())
            self.run_callbacks(
                'MTurnIdx' + name + '-Mon', orb[idx, :].ravel())
            if pln == 'Sum':
                continue
            nrb = self._csorb.nr_bpms
            dorb = orb[idx, :].ravel() - self.ref_orbs[pln][:nrb]
            self.run_callbacks(f'DeltaOrb{pln:s}Avg-Mon', _bn.nanmean(dorb))
            self.run_callbacks(f'DeltaOrb{pln:s}Std-Mon', _bn.nanstd(dorb))
            self.run_callbacks(f'DeltaOrb{pln:s}Min-Mon', _bn.nanmin(dorb))
            self.run_callbacks(f'DeltaOrb{pln:s}Max-Mon', _bn.nanmax(dorb))

    def _update_singlepass_orbits(self):
        """."""
        orbs = {'X': [], 'Y': [], 'Sum': []}
        down = self._spass_average
        with self._lock_raw_orbs:  # I need the lock here to assure consistency
            dic = {
                'maskbeg': self._spass_mask[0],
                'maskend': self._spass_mask[1],
                'nturns': down}
            nr_pts = self._smooth_npts
            for i, bpm in enumerate(self.bpms):
                dic.update({
                    'refx': self.ref_orbs['X'][i],
                    'refy': self.ref_orbs['Y'][i]})
                orbx, orby, summ = bpm.calc_sp_multiturn_pos(**dic)
                orbs['X'].append(orbx)
                orbs['Y'].append(orby)
                orbs['Sum'].append(summ)

            for pln, raw in self.raw_sporbs.items():
                norb = _np.array(orbs[pln], dtype=float).T  # turns x bpms
                norb = norb.reshape(-1)
                if self.update_raws:
                    raw.append(norb)
                del raw[:-nr_pts]
                if not raw:
                    return
                if self._smooth_meth == self._csorb.SmoothMeth.Average:
                    orb = _np.mean(raw, axis=0)
                else:
                    orb = _np.median(raw, axis=0)
                if down > 1:
                    orb = _np.mean(orb.reshape(down, -1), axis=0)
                self.smooth_sporb[pln] = orb
                name = ('Orb' if pln != 'Sum' else '') + pln
                self.run_callbacks('SPass' + name + '-Mon', list(orb))

    def _update_status(self):
        """."""
        status = 0b11111

        if self.is_trigmode():
            tim_conn = self.timing.connected
            tim_conf = self.timing.is_ok
        else:
            tim_conn = True
            tim_conf = True
        status = _util.update_bit(status, bit_pos=0, bit_val=not tim_conn)
        status = _util.update_bit(status, bit_pos=1, bit_val=not tim_conf)

        bpms = self._get_used_bpms()
        bpm_conn = all(bpm.connected for bpm in bpms)
        bpm_stt = all(bpm.state for bpm in bpms)
        status = _util.update_bit(v=status, bit_pos=2, bit_val=not bpm_conn)
        status = _util.update_bit(v=status, bit_pos=3, bit_val=not bpm_stt)

        isok = True
        if self.is_trigmode():
            isok = all(map(lambda x: x.is_ok, bpms))
        elif self.is_sloworb():
            isok = all(map(
                lambda x: x.switching_mode == _csbpm.SwModes.switching, bpms))
        status = _util.update_bit(v=status, bit_pos=4, bit_val=not isok)

        self._status = status
        self.run_callbacks('OrbStatus-Mon', status)
        self._update_bpmoffsets()

    def _update_bpmoffsets(self):
        """."""
        nrb = self._csorb.nr_bpms
        orbx = _np.zeros(nrb, dtype=float)
        orby = orbx.copy()
        for i, bpm in enumerate(self.bpms):
            orbx[i::nrb] = bpm.offsetx or 0.0
            orby[i::nrb] = bpm.offsety or 0.0
        self.run_callbacks('BPMOffsetX-Mon', orbx)
        self.run_callbacks('BPMOffsetY-Mon', orby)

    def _get_mask(self):
        if self.sofb is not None and self.sofb.matrix is not None:
            mask = self.sofb.matrix.bpm_enbllist
            mask = mask[:mask.size//2] & mask[mask.size//2:]
            mask = mask[:len(self.bpms)]
        return mask

    def _get_used_bpms(self):
        bpms = []
        mask = self._get_mask()
        for i, bpm in enumerate(self.bpms):
            bpm.put_enable = mask[i]
            if mask[i]:
                bpms.append(bpm)
        return bpms



"""SI fast orbit feedback system configuration."""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# When using this type of configuration to set the machine,
# the list of PVs should be processed in the same order they are stored
# in the configuration. The second numeric parameter in the pair is the
# delay [s] the client should wait before setting the next PV.

_corrs = [
    'SI-01M1:PS-FCH',
    'SI-01M1:PS-FCV',
    'SI-01M2:PS-FCH',
    'SI-01M2:PS-FCV',
    'SI-01C2:PS-FCH',
    'SI-01C2:PS-FCV',
    'SI-01C3:PS-FCH',
    'SI-01C3:PS-FCV',
    'SI-02M1:PS-FCH',
    'SI-02M1:PS-FCV',
    'SI-02M2:PS-FCH',
    'SI-02M2:PS-FCV',
    'SI-02C2:PS-FCH',
    'SI-02C2:PS-FCV',
    'SI-02C3:PS-FCH',
    'SI-02C3:PS-FCV',
    'SI-03M1:PS-FCH',
    'SI-03M1:PS-FCV',
    'SI-03M2:PS-FCH',
    'SI-03M2:PS-FCV',
    'SI-03C2:PS-FCH',
    'SI-03C2:PS-FCV',
    'SI-03C3:PS-FCH',
    'SI-03C3:PS-FCV',
    'SI-04M1:PS-FCH',
    'SI-04M1:PS-FCV',
    'SI-04M2:PS-FCH',
    'SI-04M2:PS-FCV',
    'SI-04C2:PS-FCH',
    'SI-04C2:PS-FCV',
    'SI-04C3:PS-FCH',
    'SI-04C3:PS-FCV',
    'SI-05M1:PS-FCH',
    'SI-05M1:PS-FCV',
    'SI-05M2:PS-FCH',
    'SI-05M2:PS-FCV',
    'SI-05C2:PS-FCH',
    'SI-05C2:PS-FCV',
    'SI-05C3:PS-FCH',
    'SI-05C3:PS-FCV',
    'SI-06M1:PS-FCH',
    'SI-06M1:PS-FCV',
    'SI-06M2:PS-FCH',
    'SI-06M2:PS-FCV',
    'SI-06C2:PS-FCH',
    'SI-06C2:PS-FCV',
    'SI-06C3:PS-FCH',
    'SI-06C3:PS-FCV',
    'SI-07M1:PS-FCH',
    'SI-07M1:PS-FCV',
    'SI-07M2:PS-FCH',
    'SI-07M2:PS-FCV',
    'SI-07C2:PS-FCH',
    'SI-07C2:PS-FCV',
    'SI-07C3:PS-FCH',
    'SI-07C3:PS-FCV',
    'SI-08M1:PS-FCH',
    'SI-08M1:PS-FCV',
    'SI-08M2:PS-FCH',
    'SI-08M2:PS-FCV',
    'SI-08C2:PS-FCH',
    'SI-08C2:PS-FCV',
    'SI-08C3:PS-FCH',
    'SI-08C3:PS-FCV',
    'SI-09M1:PS-FCH',
    'SI-09M1:PS-FCV',
    'SI-09M2:PS-FCH',
    'SI-09M2:PS-FCV',
    'SI-09C2:PS-FCH',
    'SI-09C2:PS-FCV',
    'SI-09C3:PS-FCH',
    'SI-09C3:PS-FCV',
    'SI-10M1:PS-FCH',
    'SI-10M1:PS-FCV',
    'SI-10M2:PS-FCH',
    'SI-10M2:PS-FCV',
    'SI-10C2:PS-FCH',
    'SI-10C2:PS-FCV',
    'SI-10C3:PS-FCH',
    'SI-10C3:PS-FCV',
    'SI-11M1:PS-FCH',
    'SI-11M1:PS-FCV',
    'SI-11M2:PS-FCH',
    'SI-11M2:PS-FCV',
    'SI-11C2:PS-FCH',
    'SI-11C2:PS-FCV',
    'SI-11C3:PS-FCH',
    'SI-11C3:PS-FCV',
    'SI-12M1:PS-FCH',
    'SI-12M1:PS-FCV',
    'SI-12M2:PS-FCH',
    'SI-12M2:PS-FCV',
    'SI-12C2:PS-FCH',
    'SI-12C2:PS-FCV',
    'SI-12C3:PS-FCH',
    'SI-12C3:PS-FCV',
    'SI-13M1:PS-FCH',
    'SI-13M1:PS-FCV',
    'SI-13M2:PS-FCH',
    'SI-13M2:PS-FCV',
    'SI-13C2:PS-FCH',
    'SI-13C2:PS-FCV',
    'SI-13C3:PS-FCH',
    'SI-13C3:PS-FCV',
    'SI-14M1:PS-FCH',
    'SI-14M1:PS-FCV',
    'SI-14M2:PS-FCH',
    'SI-14M2:PS-FCV',
    'SI-14C2:PS-FCH',
    'SI-14C2:PS-FCV',
    'SI-14C3:PS-FCH',
    'SI-14C3:PS-FCV',
    'SI-15M1:PS-FCH',
    'SI-15M1:PS-FCV',
    'SI-15M2:PS-FCH',
    'SI-15M2:PS-FCV',
    'SI-15C2:PS-FCH',
    'SI-15C2:PS-FCV',
    'SI-15C3:PS-FCH',
    'SI-15C3:PS-FCV',
    'SI-16M1:PS-FCH',
    'SI-16M1:PS-FCV',
    'SI-16M2:PS-FCH',
    'SI-16M2:PS-FCV',
    'SI-16C2:PS-FCH',
    'SI-16C2:PS-FCV',
    'SI-16C3:PS-FCH',
    'SI-16C3:PS-FCV',
    'SI-17M1:PS-FCH',
    'SI-17M1:PS-FCV',
    'SI-17M2:PS-FCH',
    'SI-17M2:PS-FCV',
    'SI-17C2:PS-FCH',
    'SI-17C2:PS-FCV',
    'SI-17C3:PS-FCH',
    'SI-17C3:PS-FCV',
    'SI-18M1:PS-FCH',
    'SI-18M1:PS-FCV',
    'SI-18M2:PS-FCH',
    'SI-18M2:PS-FCV',
    'SI-18C2:PS-FCH',
    'SI-18C2:PS-FCV',
    'SI-18C3:PS-FCH',
    'SI-18C3:PS-FCV',
    'SI-19M1:PS-FCH',
    'SI-19M1:PS-FCV',
    'SI-19M2:PS-FCH',
    'SI-19M2:PS-FCV',
    'SI-19C2:PS-FCH',
    'SI-19C2:PS-FCV',
    'SI-19C3:PS-FCH',
    'SI-19C3:PS-FCV',
    'SI-20M1:PS-FCH',
    'SI-20M1:PS-FCV',
    'SI-20M2:PS-FCH',
    'SI-20M2:PS-FCV',
    'SI-20C2:PS-FCH',
    'SI-20C2:PS-FCV',
    'SI-20C3:PS-FCH',
    'SI-20C3:PS-FCV',
    ]
_corr_propts = [
    [':CurrLoopKp-SP', 5000000, 0.0],         # current loop Kp parameter
    [':CurrLoopTi-SP', 300, 0.0],             # current loop Ti parameter
    [':CurrGain-SP', 6.25e-5, 0.0],           # current gain
    [':CurrOffset-SP', 0, 0.0],               # current offset
    [':VoltGain-SP', 1.12916762036e-4, 0.0],  # voltage gain
    [':VoltOffset-SP', 0, 0.0],               # voltage offset
]
_pvs_corrs = list()
for dev in _corrs:
    for ppt, val, dly in _corr_propts:
        _pvs_corrs.append([dev+ppt, val, dly])

_template_dict = {
    'pvs':
        _pvs_corrs
    }



#!/usr/bin/env python-sirius

"""Unittest module for siggen.py."""

from unittest import TestCase
import siriuspy.util as util
import siriuspy.pwrsupply.siggen as siggen


public_interface = (
    'DEFAULT_SIGGEN_CONFIG',
    'Signal',
    'SignalSine',
    'SignalDampedNSine',
    'SignalDampedSine',
    'SignalDampedSquaredSine',
    'SignalTrapezoidal',
    'SignalSquare',
    'SigGenFactory',
)


class TestModule(TestCase):
    """Test Module."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(siggen, public_interface)
        self.assertTrue(valid)

    def test_DEFAULT_SIGGEN_CONFIG(self):
        """Test DEFAULT_SIGGEN_CONFIG."""
        defconfig = siggen.DEFAULT_SIGGEN_CONFIG
        self.assertIsInstance(defconfig, tuple)
        self.assertTrue(len(defconfig), 9)


class TestSigGenConfig(TestCase):
    """Test SigGenconfig class."""

    public_interface = (
        'duration',
        'value',
        'cycle_time',
        'rampup_time',
        'theta_begin',
        'rampdown_time',
        'theta_end',
        'plateau_time',
        'decay_time',
        'get_waveform',
        'reset',
    )

    def test_public_interface(self):
        """Test class's public interface."""
        valid = util.check_public_interface_namespace(
            siggen.Signal, TestSigGenConfig.public_interface)
        self.assertTrue(valid)

    def test_duration(self):
        """Test duration."""
        # TODO: implement test!
        pass

    def test_rampup_time(self):
        """Test rampup_time."""
        # TODO: implement test!
        pass

    def test_theta_begin(self):
        """Test theta_begin."""
        # TODO: implement test!
        pass

    def test_rampdown_time(self):
        """Test rampdown_time."""
        # TODO: implement test!
        pass

    def test_theta_end(self):
        """Test theta_end."""
        # TODO: implement test!
        pass

    def test_plateau_time(self):
        """Test plateau_time."""
        # TODO: implement test!
        pass

    def test_decay_time(self):
        """Test decay_time."""
        # TODO: implement test!
        pass

    def test_get_waveform(self):
        """Test get_waveform."""
        # TODO: implement test!
        pass



"""."""

import time as _time

import numpy as _np

from ..sofb.csdev import SOFBFactory
from ..sofb.utils import si_calculate_bump as _si_calculate_bump
from .device import Device as _Device


class SOFB(_Device):
    """SOFB Device."""

    class DEVICES:
        """Devices names."""

        TB = 'TB-Glob:AP-SOFB'
        BO = 'BO-Glob:AP-SOFB'
        TS = 'TS-Glob:AP-SOFB'
        SI = 'SI-Glob:AP-SOFB'
        ALL = (TB, BO, TS, SI)

    def __new__(cls, devname):
        """."""
        # check if device exists
        if devname not in SOFB.DEVICES.ALL:
            raise NotImplementedError(devname)

        # SOFB object
        data = SOFBFactory.create(devname[:2])

        if data.acc == 'SI':
            return SISOFB(devname, data)
        elif data.isring:
            return BOSOFB(devname, data)
        return TLSOFB(devname, data)


class TLSOFB(_Device):
    """SOFB Device."""

    _properties = (
        'TrigAcqChan-Sel', 'TrigAcqChan-Sts', 'OrbStatus-Mon',
        'RespMat-SP', 'RespMat-RB', 'InvRespMat-Mon',
        'KickCH-Mon', 'KickCV-Mon',
        'DeltaKickCH-Mon', 'DeltaKickCV-Mon',
        'DeltaKickCH-SP', 'DeltaKickCV-SP',
        'MaxDeltaKickCH-RB', 'MaxDeltaKickCV-RB',
        'MaxDeltaKickCH-SP', 'MaxDeltaKickCV-SP',
        'ManCorrGainCH-SP', 'ManCorrGainCV-SP',
        'ManCorrGainCH-RB', 'ManCorrGainCV-RB',
        'RefOrbX-SP', 'RefOrbY-SP',
        'RefOrbX-RB', 'RefOrbY-RB',
        'BPMXEnblList-SP', 'BPMYEnblList-SP',
        'BPMXEnblList-RB', 'BPMYEnblList-RB',
        'CHEnblList-SP', 'CVEnblList-SP',
        'CHEnblList-RB', 'CVEnblList-RB',
        'CalcDelta-Cmd', 'ApplyDelta-Cmd', 'SmoothReset-Cmd',
        'ApplyDelta-Mon',
        'SmoothNrPts-SP', 'SmoothNrPts-RB',
        'BufferCount-Mon',
        'TrigNrSamplesPre-SP',
        'TrigNrSamplesPre-RB',
        'TrigNrSamplesPost-SP',
        'TrigNrSamplesPost-RB',
        'SPassSum-Mon', 'SPassOrbX-Mon', 'SPassOrbY-Mon',
        'MeasRespMat-Cmd', 'MeasRespMat-Mon',
        'MeasRespMatKickCH-SP', 'MeasRespMatKickCH-RB',
        'MeasRespMatKickCV-SP', 'MeasRespMatKickCV-RB',
        'MeasRespMatWait-SP', 'MeasRespMatWait-RB',
        'NrSingValues-Mon', 'MinSingValue-SP', 'MinSingValue-RB',
        'TrigAcqCtrl-Sel', 'TrigAcqCtrl-Sts', 'TrigAcqConfig-Cmd',
        'SyncBPMs-Cmd',
        )

    _default_timeout = 10  # [s]
    _default_timeout_respm = 2 * 60 * 60  # [s]
    _default_timeout_kick_apply = 2  # [s]

    def __init__(self, devname, data=None):
        """."""
        # check if device exists
        if devname not in SOFB.DEVICES.ALL:
            raise NotImplementedError(devname)

        # SOFB object
        self._data = data or SOFBFactory.create(devname[:2])
        # call base class constructor
        super().__init__(devname, properties=self._properties)

    @property
    def data(self):
        """."""
        return self._data

    @property
    def trigchannel(self):
        """."""
        return self['TrigAcqChan-Sts']

    @trigchannel.setter
    def trigchannel(self, value):
        self._enum_setter(
            'TrigAcqChan-Sel', value, self._data.TrigAcqChan)

    @property
    def respmat(self):
        """."""
        return self['RespMat-RB'].reshape(self._data.nr_bpms*2, -1)

    @respmat.setter
    def respmat(self, mat):
        """."""
        self['RespMat-SP'] = _np.array(mat).ravel()

    @property
    def invrespmat(self):
        """."""
        return self['InvRespMat-Mon'].reshape(-1, self._data.nr_bpms*2)

    @property
    def trigchannel_str(self):
        """."""
        return self._data.TrigAcqChan._fields[self['TrigAcqChan-Sts']]

    @property
    def sp_trajx(self):
        """."""
        return self['SPassOrbX-Mon']

    @property
    def sp_trajy(self):
        """."""
        return self['SPassOrbY-Mon']

    @property
    def sp_sum(self):
        """."""
        return self['SPassSum-Mon']

    @property
    def trajx(self):
        """."""
        return self.sp_trajx

    @property
    def trajy(self):
        """."""
        return self.sp_trajy

    @property
    def sum(self):
        """."""
        return self.sp_sum

    @property
    def kickch(self):
        """."""
        return self['KickCH-Mon']

    @property
    def kickcv(self):
        """."""
        return self['KickCV-Mon']

    @property
    def deltakickch(self):
        """."""
        return self['DeltaKickCH-Mon']

    @deltakickch.setter
    def deltakickch(self, value):
        """."""
        self['DeltaKickCH-SP'] = value

    @property
    def deltakickcv(self):
        """."""
        return self['DeltaKickCV-Mon']

    @deltakickcv.setter
    def deltakickcv(self, value):
        """."""
        self['DeltaKickCV-SP'] = value

    @property
    def maxdeltakickch(self):
        """."""
        return self['MaxDeltaKickCH-RB']

    @maxdeltakickch.setter
    def maxdeltakickch(self, value):
        """."""
        self['MaxDeltaKickCH-SP'] = value

    @property
    def maxdeltakickcv(self):
        """."""
        return self['MaxDeltaKickCV-RB']

    @maxdeltakickcv.setter
    def maxdeltakickcv(self, value):
        """."""
        self['MaxDeltaKickCV-SP'] = value

    @property
    def mancorrgainch(self):
        """."""
        return self['ManCorrGainCH-RB']

    @mancorrgainch.setter
    def mancorrgainch(self, value):
        """."""
        self['ManCorrGainCH-SP'] = value

    @property
    def mancorrgaincv(self):
        """."""
        return self['ManCorrGainCV-RB']

    @mancorrgaincv.setter
    def mancorrgaincv(self, value):
        """."""
        self['ManCorrGainCV-SP'] = value

    @property
    def refx(self):
        """."""
        return self['RefOrbX-RB']

    @refx.setter
    def refx(self, value):
        """."""
        self['RefOrbX-SP'] = value

    @property
    def refy(self):
        """."""
        return self['RefOrbY-RB']

    @refy.setter
    def refy(self, value):
        """."""
        self['RefOrbY-SP'] = value

    @property
    def bpmxenbl(self):
        """."""
        return _np.array(self['BPMXEnblList-RB'], dtype=bool)

    @bpmxenbl.setter
    def bpmxenbl(self, value):
        """."""
        self['BPMXEnblList-SP'] = value

    @property
    def bpmyenbl(self):
        """."""
        return _np.array(self['BPMYEnblList-RB'], dtype=bool)

    @bpmyenbl.setter
    def bpmyenbl(self, value):
        """."""
        self['BPMYEnblList-SP'] = value

    @property
    def chenbl(self):
        """."""
        return _np.array(self['CHEnblList-RB'], dtype=bool)

    @chenbl.setter
    def chenbl(self, value):
        """."""
        self['CHEnblList-SP'] = value

    @property
    def cvenbl(self):
        """."""
        return _np.array(self['CVEnblList-RB'], dtype=bool)

    @cvenbl.setter
    def cvenbl(self, value):
        """."""
        self['CVEnblList-SP'] = value

    @property
    def buffer_count(self):
        """."""
        return self['BufferCount-Mon']

    @property
    def nr_points(self):
        """."""
        return self['SmoothNrPts-RB']

    @nr_points.setter
    def nr_points(self, value):
        """."""
        self['SmoothNrPts-SP'] = int(value)

    @property
    def nr_singvals(self):
        """."""
        return self['NrSingValues-Mon']

    @property
    def singval_min(self):
        """."""
        return self['MinSingValue-RB']

    @singval_min.setter
    def singval_min(self, value):
        """."""
        self['MinSingValue-SP'] = value

    @property
    def trigacq(self):
        """."""
        return self['TrigAcqCtrl-Sts']

    @property
    def trigsamplepre(self):
        """."""
        return self['TrigNrSamplesPre-RB']

    @trigsamplepre.setter
    def trigsamplepre(self, value):
        """."""
        self['TrigNrSamplesPre-SP'] = int(value)

    @property
    def trigsamplepost(self):
        """."""
        return self['TrigNrSamplesPost-RB']

    @trigsamplepost.setter
    def trigsamplepost(self, value):
        """."""
        self['TrigNrSamplesPost-SP'] = int(value)

    def cmd_reset(self):
        """."""
        self['SmoothReset-Cmd'] = 1
        return True

    def cmd_calccorr(self):
        """."""
        self['CalcDelta-Cmd'] = 1
        return True

    def cmd_applycorr_ch(self):
        """."""
        self['ApplyDelta-Cmd'] = self._data.ApplyDelta.CH
        return True

    def cmd_applycorr_cv(self):
        """."""
        self['ApplyDelta-Cmd'] = self._data.ApplyDelta.CV
        return True

    def cmd_applycorr_rf(self):
        """."""
        self['ApplyDelta-Cmd'] = self._data.ApplyDelta.RF
        return True

    def cmd_applycorr_all(self):
        """."""
        self['ApplyDelta-Cmd'] = self._data.ApplyDelta.All
        return True

    def cmd_measrespmat_start(self):
        """."""
        self['MeasRespMat-Cmd'] = 0
        return True

    def cmd_measrespmat_stop(self):
        """."""
        self['MeasRespMat-Cmd'] = 1
        return True

    def cmd_measrespmat_reset(self):
        """."""
        self['MeasRespMat-Cmd'] = 2
        return True

    def cmd_trigacq_start(self, timeout=10):
        """."""
        self['TrigAcqCtrl-Sel'] = 'Start'
        ret = self._wait(
            'TrigAcqCtrl-Sts', self._data.TrigAcqCtrl.Start, timeout=timeout)
        if not ret:
            return False
        _time.sleep(0.6)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    def cmd_trigacq_stop(self, timeout=10):
        """."""
        self['TrigAcqCtrl-Sel'] = 'Stop'
        ret = self._wait(
            'TrigAcqCtrl-Sts', self._data.TrigAcqCtrl.Stop, timeout=timeout)
        if not ret:
            return False
        _time.sleep(0.6)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    def cmd_trigacq_abort(self, timeout=10):
        """."""
        self['TrigAcqCtrl-Sel'] = 'Abort'
        ret = self._wait(
            'TrigAcqCtrl-Sts', self._data.TrigAcqCtrl.Abort, timeout=timeout)
        if not ret:
            return False
        _time.sleep(0.6)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    def cmd_trigacq_config(self, timeout=10):
        """."""
        self['TrigAcqConfig-Cmd'] = 1
        return self.wait_orb_status_ok(timeout=timeout)

    @property
    def applydeltakick_mon(self):
        """."""
        return self['ApplyDelta-Mon']

    @property
    def measrespmat_mon(self):
        """."""
        return self['MeasRespMat-Mon']

    @property
    def measrespmat_kickch(self):
        """."""
        return self['MeasRespMatKickCH-RB']

    @measrespmat_kickch.setter
    def measrespmat_kickch(self, value):
        self['MeasRespMatKickCH-SP'] = value

    @property
    def measrespmat_kickcv(self):
        """."""
        return self['MeasRespMatKickCV-RB']

    @measrespmat_kickcv.setter
    def measrespmat_kickcv(self, value):
        self['MeasRespMatKickCV-SP'] = value

    @property
    def measrespmat_wait(self):
        """."""
        return self['MeasRespMatWait-RB']

    @measrespmat_wait.setter
    def measrespmat_wait(self, value):
        self['MeasRespMatWait-SP'] = value

    def correct_orbit_manually(self, nr_iters=10, residue=5):
        """."""
        self.cmd_turn_off_autocorr()
        for i in range(nr_iters):
            resx = self.orbx - self.refx
            resy = self.orby - self.refy
            resx = resx[self.bpmxenbl.nonzero()[0]]
            resy = resy[self.bpmyenbl.nonzero()[0]]
            resx = _np.sqrt(_np.sum(resx*resx)/resx.size)
            resy = _np.sqrt(_np.sum(resy*resy)/resy.size)
            if resx < residue and resy < residue:
                break
            self.cmd_calccorr()
            _time.sleep(0.5)
            self.cmd_applycorr_all()
            self.wait_apply_delta_kick()
            _time.sleep(0.2)
            self.cmd_reset()
            self.wait_buffer()
        return i, resx, resy

    def wait_buffer(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        return self._wait(
            'BufferCount-Mon', self.nr_points, timeout=timeout, comp='ge')

    def wait_apply_delta_kick(self, timeout=None):
        """."""
        def_timeout = min(1.05*self.deltakickrf, self.maxdeltakickrf) // 20
        def_timeout = max(self._default_timeout_kick_apply, def_timeout)
        timeout = timeout or def_timeout
        return self._wait(
            'ApplyDelta-Mon', self._data.ApplyDeltaMon.Applying,
            timeout=timeout, comp='ne')

    def wait_respm_meas(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout_respm
        return self._wait(
            'MeasRespMat-Mon', self._data.MeasRespMatMon.Measuring,
            timeout=timeout, comp='ne')

    def wait_orb_status_ok(self, timeout=10):
        """."""
        return self._wait('OrbStatus-Mon', 0, timeout=timeout)

    def cmd_sync_bpms(self):
        """Synchronize BPMs."""
        self['SyncBPMs-Cmd'] = 1
        return True


class BOSOFB(TLSOFB):
    """SOFB Device."""

    _properties = TLSOFB._properties + (
        'MTurnAcquire-Cmd',
        'MTurnSum-Mon', 'MTurnOrbX-Mon', 'MTurnOrbY-Mon',
        'MTurnIdxOrbX-Mon', 'MTurnIdxOrbY-Mon', 'MTurnIdxSum-Mon',
        'MTurnTime-Mon',
        )

    @property
    def mt_trajx(self):
        """."""
        return self['MTurnOrbX-Mon']

    @property
    def mt_trajy(self):
        """."""
        return self['MTurnOrbY-Mon']

    @property
    def mt_sum(self):
        """."""
        return self['MTurnSum-Mon']

    @property
    def mt_time(self):
        """."""
        return self['MTurnTime-Mon']

    @property
    def mt_trajx_idx(self):
        """."""
        return self['MTurnIdxOrbX-Mon']

    @property
    def mt_trajy_idx(self):
        """."""
        return self['MTurnIdxOrbY-Mon']

    @property
    def mt_sum_idx(self):
        """."""
        return self['MTurnIdxSum-Mon']

    @property
    def trajx(self):
        """."""
        return self.mt_trajx_idx

    @property
    def trajy(self):
        """."""
        return self.mt_trajy_idx

    @property
    def sum(self):
        """."""
        return self.mt_sum_idx

    def cmd_mturn_acquire(self):
        """."""
        self['MTurnAcquire-Cmd'] = 1
        return True


class SISOFB(BOSOFB):
    """SOFB Device."""

    _properties = BOSOFB._properties + (
        'SOFBMode-Sel', 'SOFBMode-Sts',
        'KickRF-Mon',
        'DeltaKickRF-Mon', 'DeltaKickRF-SP',
        'MaxDeltaKickRF-RB', 'MaxDeltaKickRF-SP',
        'ManCorrGainRF-SP', 'ManCorrGainRF-RB',
        'MeasRespMatKickRF-SP', 'MeasRespMatKickRF-RB',
        'RFEnbl-Sel', 'RFEnbl-Sts',
        'SlowOrbX-Mon', 'SlowOrbY-Mon',
        'LoopState-Sts', 'LoopState-Sel',
        'DriveFreqDivisor-SP', 'DriveFreqDivisor-RB', 'DriveFrequency-Mon',
        'DriveNrCycles-SP', 'DriveNrCycles-RB', 'DriveDuration-Mon',
        'DriveAmplitude-SP', 'DriveAmplitude-RB',
        'DrivePhase-SP', 'DrivePhase-RB',
        'DriveCorrIndex-SP', 'DriveCorrIndex-RB',
        'DriveBPMIndex-SP', 'DriveBPMIndex-RB',
        'DriveType-Sel', 'DriveType-Sts',
        'DriveState-Sel', 'DriveState-Sts',
        'DriveData-Mon',
        )

    @property
    def opmode(self):
        """."""
        return self['SOFBMode-Sts']

    @opmode.setter
    def opmode(self, value):
        self._enum_setter(
            'SOFBMode-Sel', value, self._data.SOFBMode)

    @property
    def opmode_str(self):
        """."""
        return self._data.SOFBMode._fields[self['SOFBMode-Sts']]

    def cmd_change_opmode_to_multiturn(self, timeout=10):
        """."""
        mode = self._data.SOFBMode.MultiTurn
        self.opmode = mode
        ret = self._wait('SOFBMode-Sts', mode, timeout=timeout)
        if not ret:
            return False
        _time.sleep(1)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    def cmd_change_opmode_to_sloworb(self, timeout=10):
        """."""
        mode = self._data.SOFBMode.SlowOrb
        self.opmode = mode
        ret = self._wait('SOFBMode-Sts', mode, timeout=timeout)
        if not ret:
            return False
        _time.sleep(0.6)  # Status PV updates at 2Hz
        return self.wait_orb_status_ok(timeout=timeout)

    @property
    def drivests(self):
        """."""
        return self['DriveState-Sts']

    @property
    def drivetype(self):
        """."""
        return self['DriveType-Sts']

    @drivetype.setter
    def drivetype(self, value):
        self._enum_setter(
            'DriveType-Sel', value, self._data.DriveType)

    @property
    def drivetype_str(self):
        """."""
        return self._data.DriveType._fields[self['DriveType-Sts']]

    @property
    def drivefreqdivisor(self):
        """."""
        return self['DriveFreqDivisor-RB']

    @drivefreqdivisor.setter
    def drivefreqdivisor(self, value):
        """."""
        self['DriveFreqDivisor-SP'] = value

    @property
    def drivefrequency_mon(self):
        """."""
        return self['DriveFrequency-Mon']

    @property
    def drivenrcycles(self):
        """."""
        return self['DriveNrCycles-RB']

    @drivenrcycles.setter
    def drivenrcycles(self, value):
        """."""
        self['DriveNrCycles-SP'] = value

    @property
    def driveduration_mon(self):
        """."""
        return self['DriveDuration-Mon']

    @property
    def driveamplitude(self):
        """."""
        return self['DriveAmplitude-RB']

    @driveamplitude.setter
    def driveamplitude(self, value):
        """."""
        self['DriveAmplitude-SP'] = value

    @property
    def drivephase(self):
        """."""
        return self['DrivePhase-RB']

    @drivephase.setter
    def drivephase(self, value):
        """."""
        self['DrivePhase-SP'] = value

    @property
    def drivecorridx(self):
        """."""
        return self['DriveCorrIndex-RB']

    @drivecorridx.setter
    def drivecorridx(self, value):
        """."""
        self['DriveCorrIndex-SP'] = value

    @property
    def drivebpmidx(self):
        """."""
        return self['DriveBPMIndex-RB']

    @drivebpmidx.setter
    def drivebpmidx(self, value):
        """."""
        self['DriveBPMIndex-SP'] = value

    @property
    def drivedata_time(self):
        """."""
        return self['DriveData-Mon'][::3]

    @property
    def drivedata_corr(self):
        """."""
        return self['DriveData-Mon'][1::3]

    @property
    def drivedata_bpm(self):
        """."""
        return self['DriveData-Mon'][2::3]

    @property
    def orbx(self):
        """."""
        return self['SlowOrbX-Mon'] if self._data.isring else None

    @property
    def orby(self):
        """."""
        return self['SlowOrbY-Mon'] if self._data.isring else None

    @property
    def kickrf(self):
        """."""
        return self['KickRF-Mon']

    @property
    def deltakickrf(self):
        """."""
        return self['DeltaKickRF-Mon']

    @deltakickrf.setter
    def deltakickrf(self, value):
        """."""
        self['DeltaKickRF-SP'] = value

    @property
    def maxdeltakickrf(self):
        """."""
        return self['MaxDeltaKickRF-RB']

    @maxdeltakickrf.setter
    def maxdeltakickrf(self, value):
        """."""
        self['MaxDeltaKickRF-SP'] = value

    @property
    def mancorrgainrf(self):
        """."""
        return self['ManCorrGainRF-RB']

    @mancorrgainrf.setter
    def mancorrgainrf(self, value):
        """."""
        self['ManCorrGainRF-SP'] = value

    @property
    def rfenbl(self):
        """."""
        if self._data.acc_idx == self._data.Rings.SI:
            return bool(self['RFEnbl-Sts'])

    @rfenbl.setter
    def rfenbl(self, value):
        """."""
        if self._data.acc_idx == self._data.Rings.SI:
            self['RFEnbl-Sel'] = value

    @property
    def measrespmat_kickrf(self):
        """."""
        return self['MeasRespMatKickRF-RB']

    @measrespmat_kickrf.setter
    def measrespmat_kickrf(self, value):
        self['MeasRespMatKickRF-SP'] = value

    @property
    def autocorrsts(self):
        """."""
        return self['LoopState-Sts']

    def cmd_turn_on_autocorr(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        if self.autocorrsts == self._data.LoopState.Closed:
            return True
        self['LoopState-Sel'] = self._data.LoopState.Closed
        return self._wait(
            'LoopState-Sts', self._data.LoopState.Closed, timeout=timeout)

    def cmd_turn_off_autocorr(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        if self.autocorrsts == self._data.LoopState.Open:
            return True
        self['LoopState-Sel'] = self._data.LoopState.Open
        return self._wait(
            'LoopState-Sts', self._data.LoopState.Open, timeout=timeout)

    def cmd_turn_on_drive(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        if self.drivests == self._data.DriveState.Closed:
            return True
        self['DriveState-Sel'] = self._data.DriveState.Closed
        return self._wait(
            'DriveState-Sts', self._data.DriveState.Closed, timeout=timeout)

    def cmd_turn_off_drive(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        if self.drivests == self._data.DriveState.Open:
            return True
        self['DriveState-Sel'] = self._data.DriveState.Open
        return self.wait_drive(timeout=timeout)

    def wait_drive(self, timeout=None):
        """."""
        timeout = timeout or self._default_timeout_respm
        return self._wait(
            'DriveState-Sts', self._data.DriveState.Open, timeout=timeout)

    @staticmethod
    def si_calculate_bumps(orbx, orby, subsec, agx=0, agy=0, psx=0, psy=0):
        """."""
        return _si_calculate_bump(
            orbx, orby, subsec, agx=agx, agy=agy, psx=psx, psy=psy)

    @property
    def trajx(self):
        """."""
        if self.opmode == self._data.SOFBMode.MultiTurn:
            return self.mt_trajx_idx
        return self.sp_trajx

    @property
    def trajy(self):
        """."""
        if self.opmode == self._data.SOFBMode.MultiTurn:
            return self.mt_trajy_idx
        return self.sp_trajy

    @property
    def sum(self):
        """."""
        if self.opmode == self._data.SOFBMode.MultiTurn:
            return self.mt_sum_idx
        return self.sp_sum



"""Main Module of the IOC Logic."""

import warnings
import time as _time
import numpy as _np
from epics import PV as _PV

from ...callbacks import Callback as _Callback
from ...epics import SiriusPVTimeSerie as _SiriusPVTimeSerie
from ...envars import VACA_PREFIX as _vaca_prefix
from ..csdev import \
    Const as _Const, get_lifetime_database as _get_database

_MAX_BUFFER_SIZE = 36000


class SILifetimeApp(_Callback):
    """Main Class of the IOC Logic."""

    def __init__(self):
        """Class constructor."""
        super().__init__()

        self._init_time = _time.time()

        self._pvs_database = _get_database()
        self._mode = _Const.Fit.Linear
        self._current_offset = 0.0
        self._frst_smpl_ts = self._init_time
        self._last_smpl_ts = -1
        self._frst_smpl_ts_dcct = self._init_time
        self._last_smpl_ts_dcct = -1
        self._smpl_intvl_mon_dcct = 0.0
        self._frst_smpl_ts_bpm = self._init_time
        self._last_smpl_ts_bpm = -1
        self._smpl_intvl_mon_bpm = 0.0
        self._last_ts_set = 'last'
        self._sampling_interval = 500.0
        self._min_intvl_btw_spl = 0.0
        self._buffautorst_mode = _Const.BuffAutoRst.DCurrCheck
        self._buffautorst_dcurr = 0.01
        self._is_stored = 0
        self._lifetime = 0
        self._lifetime_bpm = 0

        self._prefix = _vaca_prefix + ('-' if _vaca_prefix else '')
        self._current_pv = _PV(
            self._prefix+'SI-Glob:AP-CurrInfo:Current-Mon',
            connection_timeout=0.05)
        self._bpmsum_pv = _PV(
            self._prefix+'SI-01M1:DI-BPM:Sum-Mon',
            connection_timeout=0.05)
        self._storedebeam_pv = _PV(
            self._prefix+'SI-Glob:AP-CurrInfo:StoredEBeam-Mon',
            connection_timeout=0.05)

        self._current_buffer = _SiriusPVTimeSerie(
            pv=self._current_pv, mode=0, nr_max_points=_MAX_BUFFER_SIZE,
            use_pv_timestamp=False)
        self._bpmsum_buffer = _SiriusPVTimeSerie(
            pv=self._bpmsum_pv, mode=0, nr_max_points=_MAX_BUFFER_SIZE,
            use_pv_timestamp=False)

        self._current_pv.add_callback(self._callback_calclifetime)
        self._bpmsum_pv.add_callback(self._callback_calclifetime)
        self._storedebeam_pv.add_callback(self._callback_get_storedebeam)

    @property
    def pvs_database(self):
        """Return pvs database."""
        return self._pvs_database

    def init_database(self):
        """Set PVs initial values."""
        self.run_callbacks('FrstSplTime-SP', self._frst_smpl_ts)
        self.run_callbacks('LastSplTime-SP', self._last_smpl_ts)
        self.run_callbacks('FrstSplTime-RB', self._frst_smpl_ts)
        self.run_callbacks('LastSplTime-RB', self._last_smpl_ts)
        self.run_callbacks('FrstSplTimeBPM-RB', self._frst_smpl_ts_bpm)
        self.run_callbacks('LastSplTimeBPM-RB', self._last_smpl_ts_bpm)

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def read(self, reason):
        """Read from IOC database."""
        value = None
        if reason in ['Lifetime-Mon', 'LifetimeBPM-Mon']:
            is_bpm = 'BPM' in reason
            lt_type = 'BPM' if is_bpm else ''
            lt_name = '_lifetime'+('_bpm' if is_bpm else '')
            buffer_dt = self._bpmsum_buffer if is_bpm else self._current_buffer

            # get first and last sample
            now = _time.time()
            self._update_times(now)
            first_name = '_frst_smpl_ts'+('_bpm' if is_bpm else '_dcct')
            first_smpl = getattr(self, first_name)
            last_name = '_last_smpl_ts'+('_bpm' if is_bpm else '_dcct')
            last_smpl = getattr(self, last_name)
            last_smpl = now if last_smpl == -1 else min(last_smpl, now)
            intvl_name = '_smpl_intvl_mon'+('_bpm' if is_bpm else '_dcct')
            intvl_smpl = getattr(self, intvl_name)

            # calculate lifetime
            ts_abs_dqorg, val_dqorg = buffer_dt.get_serie(time_absolute=True)
            ts_dqorg = ts_abs_dqorg - now
            ts_dq, val_dq, ts_abs_dq = self._filter_buffer(
                ts_dqorg, val_dqorg, ts_abs_dqorg, first_smpl, last_smpl)

            if ts_dq.size != 0:
                if first_smpl != ts_abs_dq[0]:
                    setattr(self, first_name, ts_abs_dq[0])
                    self.run_callbacks(
                        'FrstSplTime'+lt_type+'-RB', getattr(self, first_name))
                intvl_smpl = last_smpl - first_smpl
                setattr(self, intvl_name, intvl_smpl)
                self.run_callbacks(
                    'SplIntvl'+lt_type+'-Mon', getattr(self, intvl_name))

                val_dq -= self._current_offset

                # check min number of points in buffer
                if len(val_dq) > 100:
                    fit = 'lin' if self._mode == _Const.Fit.Linear else 'exp'
                    value = self._least_squares_fit(ts_dq, val_dq, fit=fit)
                    setattr(self, lt_name, value)

            # update pvs
            self.run_callbacks('BufferValue'+lt_type+'-Mon', val_dq)
            self.run_callbacks('BufferTimestamp'+lt_type+'-Mon', ts_dq)
            self.run_callbacks('BuffSize'+lt_type+'-Mon', len(val_dq))
            self.run_callbacks('BuffSizeTot'+lt_type+'-Mon', len(val_dqorg))
        return value

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        status = False
        now = _time.time()
        if reason == 'MaxSplIntvl-SP':
            if value == -1 or value >= 0:
                self._sampling_interval = value
                if value != -1:
                    self._update_times(now, force_min_first=True)
            elif value < 0:
                self._sampling_interval = 0
            self.run_callbacks('MaxSplIntvl-RB', self._sampling_interval)
            status = True
        elif reason == 'MinIntvlBtwSpl-SP':
            if value < 0:
                value = 0
            self._min_intvl_btw_spl = value
            self.run_callbacks('MinIntvlBtwSpl-RB', value)
            status = True
        elif reason == 'LtFitMode-Sel':
            self._mode = value
            self.run_callbacks('LtFitMode-Sts', value)
            status = True
        elif reason == 'CurrOffset-SP':
            self._current_offset = value
            self.run_callbacks('CurrOffset-RB', value)
            status = True
        elif reason == 'BuffRst-Cmd':
            self._reset_buff()
            status = True
        elif reason == 'BuffAutoRst-Sel':
            self._buffautorst_mode = value
            self.run_callbacks('BuffAutoRst-Sts', value)
            status = True
        elif reason == 'BuffAutoRstDCurr-SP':
            self._buffautorst_dcurr = value
            self.run_callbacks('BuffAutoRstDCurr-RB', value)
            status = True
        elif reason == 'FrstSplTime-SP':
            self._frst_smpl_ts = value
            self._frst_smpl_ts_dcct = value
            self._frst_smpl_ts_bpm = value
            self._last_ts_set = 'first'
            self._update_times(now)
            status = True
        elif reason == 'LastSplTime-SP':
            self._last_smpl_ts = value
            self._last_smpl_ts_dcct = value
            self._last_smpl_ts_bpm = value
            self._last_ts_set = 'last'
            self._update_times(now, force_min_first=True)
            status = True
        return status

    # ---------- callbacks ----------

    def _callback_get_storedebeam(self, value, **kws):
        self._is_stored = value

    def _callback_calclifetime(self, pvname, timestamp, **kws):
        # check DCCT StoredEBeam PV
        if not self._is_stored:
            self._buffautorst_check()
            return

        is_bpm = 'BPM' in pvname
        buffer_dt = self._bpmsum_buffer if is_bpm else self._current_buffer

        # try to add a new point to buffer
        if not buffer_dt.acquire():
            return

        # check whether the buffer must be reset
        self._buffautorst_check()

    # ---------- auxiliar methods ----------

    def _buffautorst_check(self):
        """Check situations to clear internal buffer.
        If BuffAutoRst == DCurrCheck, check abrupt variation of current.
        """
        if self._buffautorst_mode == _Const.BuffAutoRst.Off:
            return

        reset = False
        [_, value] = self._current_buffer.serie
        if len(value) >= 2:
            deltacurr = abs(value[-1] - value[-2])
        else:
            deltacurr = 0.0
        if deltacurr > self._buffautorst_dcurr:
            reset = True
        if reset:
            self._reset_buff()

    def _reset_buff(self):
        now = _time.time()
        self._frst_smpl_ts = now
        self._last_smpl_ts = -1
        self._frst_smpl_ts_dcct = now
        self._last_smpl_ts_dcct = -1
        self._frst_smpl_ts_bpm = now
        self._last_smpl_ts_bpm = -1
        self.run_callbacks('FrstSplTime-RB', self._frst_smpl_ts_dcct)
        self.run_callbacks('LastSplTime-RB', self._last_smpl_ts_dcct)
        self.run_callbacks('FrstSplTimeBPM-RB', self._frst_smpl_ts_bpm)
        self.run_callbacks('LastSplTimeBPM-RB', self._last_smpl_ts_bpm)

    def _filter_buffer(self, timestamp, value, abs_timestamp, first, last):
        ts_arrayorg = _np.asarray(timestamp)
        val_arrayorg = _np.asarray(value)
        ts_abs_arrayorg = _np.asarray(abs_timestamp)
        if ts_abs_arrayorg.size == 0:
            return ts_arrayorg, val_arrayorg, ts_abs_arrayorg

        ind1 = ts_abs_arrayorg >= first
        ind2 = ts_abs_arrayorg <= last
        indices = _np.logical_and(ind1, ind2).nonzero()
        if not indices[0].size:
            return _np.array([]), _np.array([]), _np.array([])
        slc = slice(indices[0][0], indices[0][-1]+1)

        ts_aux_array = ts_arrayorg[slc]
        val_aux_array = val_arrayorg[slc]
        ts_aux_abs_array = ts_abs_arrayorg[slc]

        min_intvl = self._min_intvl_btw_spl
        if min_intvl <= 0.0:
            return ts_aux_array, val_aux_array, ts_aux_abs_array

        indices = _np.zeros(len(ts_aux_array), dtype=bool)
        indices[0] = _np.True_
        last_idx = 0
        for i, _ in enumerate(ts_aux_array):
            if abs(ts_aux_array[last_idx] - ts_aux_array[i]) > min_intvl:
                indices[i] = _np.True_
                last_idx = i

        indices = indices.nonzero()
        ts_array = ts_aux_array[indices]
        val_array = val_aux_array[indices]
        ts_abs_array = ts_aux_abs_array[indices]
        return ts_array, val_array, ts_abs_array

    @staticmethod
    def _least_squares_fit(timestamp, value, fit='exp'):
        if fit == 'exp':
            with warnings.catch_warnings():
                warnings.filterwarnings('error')
                try:
                    value = _np.log(value)
                except Warning:
                    return 0.0
        _ns = len(timestamp)
        _x1 = _np.sum(timestamp)
        _y1 = _np.sum(value)
        if timestamp.size > 10000:
            _x2 = _np.sum(timestamp*timestamp)
            _xy = _np.sum(timestamp*value)
        else:
            _x2 = _np.dot(timestamp, timestamp)
            _xy = _np.dot(timestamp, value)
        fit_a = (_x2*_y1 - _xy*_x1)/(_ns*_x2 - _x1*_x1)
        fit_b = (_ns*_xy - _x1*_y1)/(_ns*_x2 - _x1*_x1)

        if fit == 'exp':
            lifetime = - 1/fit_b
        else:
            lifetime = - fit_a/fit_b
        return lifetime

    def _update_times(self, now, force_min_first=False):
        if self._last_ts_set == 'first':
            value = self._frst_smpl_ts
            if self._sampling_interval != -1:
                last = value + self._sampling_interval
                self._last_smpl_ts_dcct = last
                self._last_smpl_ts_bpm = last
        elif self._last_ts_set == 'last':
            value = now if self._last_smpl_ts == -1 else self._last_smpl_ts
            if self._sampling_interval != -1:
                first = value - self._sampling_interval
                if self._frst_smpl_ts_dcct < first or force_min_first:
                    self._frst_smpl_ts_dcct = first
                if self._frst_smpl_ts_bpm < first or force_min_first:
                    self._frst_smpl_ts_bpm = first

        self.run_callbacks('FrstSplTime-RB', self._frst_smpl_ts_dcct)
        self.run_callbacks('FrstSplTimeBPM-RB', self._frst_smpl_ts_bpm)
        self.run_callbacks('LastSplTime-RB', self._last_smpl_ts_dcct)
        self.run_callbacks('LastSplTimeBPM-RB', self._last_smpl_ts_bpm)



"""Define ICT device high level PVs database.

This is not a primary source database. Primary sources can be found in:
 - Digital Multimeter: https://github.com/lnls-dig/dmm7510-epics-ioc
"""

from ... import csdev as _csdev


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    RANGESEL = ('40 nC', '20 nC', '10 nC', '8 nC', '4 nC', '2 nC', '0.8 nC')
    RANGESTS = ('UNKNOWN', '40 nC', '20 nC', '10 nC', '8 nC', '4 nC', '2 nC',
                '0.8 nC')
    IMPED = ('AUTO', '10MOhm')
    SAMPLETRGSEL = ('None', 'External', 'InLevel')
    SAMPLETRGSTS = ('Unknown', 'External', 'InLevel', 'None')
    CALCHARGE = ('1 nC', '100 pC', '10 pC', '1 pC')


_et = ETypes  # syntactic sugar


# --- Database ---

def get_ict_database():
    """Return ICT device database."""
    pvs_database = {
        # Charge Measurement
        'Charge-Mon':      {'type': 'float', 'unit': 'nC', 'prec': 6},
        'ChargeHstr-Mon':  {'type': 'float', 'count': 1000, 'unit': 'nC'},

        # Measurement Reliability
        'ReliableMeas-Mon': {'type': 'int', 'hihi': 7, 'high': 1, 'hilim': 7,
                             'lolim': 0, 'low': 0, 'lolo': 0},
        'ReliableMeasLabels-Cte': {'type': 'string', 'count': 3},

        # Digitize Trigger source
        'SampleTrg-Sel': {'type': 'enum', 'enums': _et.SAMPLETRGSEL},
        'SampleTrg-Sts': {'type': 'enum', 'enums': _et.SAMPLETRGSTS},

        # Level Threshold
        'Threshold-SP': {'type': 'float', 'unit': 'nC', 'prec': 6,
                         'hihi': 40, 'high': 40, 'hilim': 40,
                         'lolim': 0, 'low': 0, 'lolo': 0},
        'Threshold-RB': {'type': 'float', 'unit': 'nC', 'prec': 6,
                         'hihi': 40, 'high': 40, 'hilim': 40,
                         'lolim': 0, 'low': 0, 'lolo': 0},
        'HFReject-Sel': {'type': 'enum', 'enums': _et.OFF_ON},
        'HFReject-Sts': {'type': 'enum', 'enums': _et.OFF_ON},

        # Measure timer configuration
        '2ndReadDly-SP': {'type': 'float', 'unit': 's', 'prec': 6,
                          'hihi': 100000, 'high': 100000,
                          'hilim': 100000, 'lolim': 0.000008,
                          'low': 0.000008, 'lolo': 0.000008},
        '2ndReadDly-RB': {'type': 'float', 'unit': 's', 'prec': 6,
                          'hihi': 100000, 'high': 100000,
                          'hilim': 100000, 'lolim': 0.000008,
                          'low': 0.000008, 'lolo': 0.000008},

        # Raw Charge Measurement
        'RawReadings-Mon': {'type': 'float', 'count': 1000, 'unit': 'nC'},
        'RawPulse-Mon':    {'type': 'float', 'count': 1000, 'unit': 'nC'},
        'RawNoise-Mon':    {'type': 'float', 'count': 1000, 'unit': 'nC'},

        # Digitize Settings
        'SampleCnt-SP':  {'type': 'int',
                          'hihi': 1000000, 'high': 1000000, 'hilim': 1000000,
                          'lolim': 1, 'low': 1, 'lolo': 1},
        'SampleCnt-RB':  {'type': 'int',
                          'hihi': 1000000, 'high': 1000000, 'hilim': 1000000,
                          'lolim': 1, 'low': 1, 'lolo': 1},
        'Aperture-SP':   {'type': 'int', 'unit': 'us', 'value': 1,
                          'hihi': 1000, 'high': 1000, 'hilim': 1000,
                          'lolim': 1, 'low': 1, 'lolo': 1},
        'Aperture-RB':   {'type': 'int', 'unit': 'us', 'value': 1,
                          'hihi': 1000, 'high': 1000, 'hilim': 1000,
                          'lolim': 1, 'low': 1, 'lolo': 1},
        'SampleRate-SP': {'type': 'int', 'unit': 'readings/s',
                          'hihi': 1000000, 'high': 1000000, 'hilim': 1000000,
                          'lolim': 1000, 'low': 1000, 'lolo': 1000},
        'SampleRate-RB': {'type': 'int', 'unit': 'readings/s',
                          'hihi': 1000000, 'high': 1000000, 'hilim': 1000000,
                          'lolim': 1000, 'low': 1000, 'lolo': 1000},
        'Imped-Sel':     {'type': 'enum', 'enums': _et.IMPED, 'value': 0},
        'Imped-Sts':     {'type': 'enum', 'enums': _et.IMPED, 'value': 0},

        # Charge Digitize Range
        'BCMRange-SP': {'type': 'float', 'unit': 'V', 'prec': 6,
                        'hihi': 10, 'high': 10, 'hilim': 10,
                        'lolim': 1, 'low': 1, 'lolo': 1},
        'Range-Sel':   {'type': 'enum', 'enums': _et.RANGESEL},
        'Range-Sts':   {'type': 'enum', 'enums': _et.RANGESTS},

        # BCM Calibration Settings
        'CalEnbl-Sel':   {'type': 'enum', 'enums': _et.OFF_ON},
        'CalEnbl-Sts':   {'type': 'enum', 'enums': _et.OFF_ON},
        'CalCharge-Sel': {'type': 'enum', 'enums': _et.CALCHARGE, 'value': 0},
        'CalCharge-Sts': {'type': 'enum', 'enums': _et.CALCHARGE, 'value': 0},

        # Multimeter Setup
        'Download-Cmd': {'type': 'int', 'value': 0}
    }
    return pvs_database



"""Define Power Supply Property and Strength Devices."""

from ..namesys import SiriusPVName as _SiriusPVName
from ..magnet.factory import NormalizerFactory as _NormalizerFactory

from .device import Devices as _Devices
from .syncd import DevicesSync as _DevicesSync


class PSProperty(_DevicesSync):
    """Power Supply Epics Connector."""

    _ps2devs = {
        # devices which are mapped to more than one device
        'BO-Fam:PS-B-1': ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2'),
        'BO-Fam:PS-B-2': ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2'),
        'SI-Fam:PS-B1B2-1': ('SI-Fam:PS-B1B2-1', 'SI-Fam:PS-B1B2-2'),
        'SI-Fam:PS-B1B2-2': ('SI-Fam:PS-B1B2-1', 'SI-Fam:PS-B1B2-2'),
        }

    def __init__(self, devname, propty, auto_mon=False):
        """."""
        devname = _SiriusPVName(devname)

        # get devnames
        devnames = PSProperty._get_devnames(devname)

        # call base class constructor
        super().__init__(
            devnames=devnames, propty_sync=[propty], auto_mon=auto_mon)

    @property
    def property_sync(self):
        """Return sole synchonized property name of device."""
        return self.properties_sync[0]

    @property
    def value(self):
        """Return property value."""
        return self.value_get(self.property_sync)

    @value.setter
    def value(self, current):
        """Set property value."""
        self.value_set(self.property_sync, current)

    @property
    def limits(self):
        """Return Property limits."""
        # NOTE: improve code.
        try:
            pvname = self.pvnames[0]
        except TypeError:
            # is a set
            for pvn in self.pvnames:
                if pvn.endswith('-SP'):
                    pvname = pvn
        pvobj = self.pv_object(pvname)
        limits = (
            pvobj.lower_alarm_limit,
            pvobj.lower_warning_limit,
            pvobj.lower_disp_limit,
            pvobj.lower_ctrl_limit,
            pvobj.upper_ctrl_limit,
            pvobj.upper_disp_limit,
            pvobj.upper_warning_limit,
            pvobj.upper_alarm_limit)
        return limits

    @staticmethod
    def _get_devnames(devname):
        """."""
        if devname in PSProperty._ps2devs:
            return PSProperty._ps2devs[devname]
        return (devname, )


class StrengthConv(_Devices):
    """Strength Converter."""

    # TODO: Test changing default value of auto_mon to see if conversion
    # IOCs are improved.

    def __init__(self, devname, proptype, auto_mon=False):
        """."""
        devname = _SiriusPVName(devname)

        # get pwrsupply normalized
        self._norm_mag = self._create_normalizer(devname)

        # get devices that provide normalization current for strengths
        self._dev_dip, self._dev_fam = \
            self._get_devices(devname, proptype, auto_mon)
        if self._dev_fam:
            devices = (self._dev_dip, self._dev_fam)
        elif self._dev_dip:
            devices = (self._dev_dip, )
        else:
            devices = ()

        # call base class constructor
        super().__init__(devname, devices)

    @property
    def dipole_strength(self):
        """Return dipole strength."""
        if self._dev_dip:
            return self._dev_dip.value
        return None

    @property
    def family_strength(self):
        """Return family strength."""
        if self._dev_fam:
            return self._dev_fam.value
        return None

    def conv_current_2_strength(
            self, currents,
            strengths_dipole=None, strengths_family=None):
        """Convert currents to strengths."""
        norm = self._norm_mag
        kwargs = self._get_kwargs(strengths_dipole, strengths_family)
        if kwargs is None:
            return None
        strengths = \
            norm.conv_current_2_strength(currents=currents, **kwargs)
        return strengths

    def conv_strength_2_current(self, strengths,
                                strengths_dipole=None,
                                strengths_family=None):
        """Convert strengths to currents."""
        norm = self._norm_mag
        kwargs = self._get_kwargs(strengths_dipole, strengths_family)
        if kwargs is None:
            return None
        currents = \
            norm.conv_strength_2_current(strengths=strengths, **kwargs)
        return currents

    def _get_kwargs(self,
                    strengths_dipole=None,
                    strengths_family=None):
        if not self._dev_dip:
            # is a dipole
            kwargs = dict()
            return kwargs
        if not self._dev_fam:
            # is not a trim
            if strengths_dipole is None:
                strengths_dipole = self._dev_dip.value
                if strengths_dipole is None:
                    return None
            kwargs = {'strengths_dipole': strengths_dipole}
            return kwargs
        # is a trim
        if strengths_dipole is None:
            strengths_dipole = self._dev_dip.value
            if strengths_dipole is None:
                return None
        if strengths_family is None:
            strengths_family = self._dev_fam.value
            if strengths_family is None:
                return None
        kwargs = {
            'strengths_dipole': strengths_dipole,
            'strengths_family': strengths_family}
        return kwargs

    @staticmethod
    def _create_normalizer(psname):
        if psname.sec == 'TB' and psname.dev == 'B':
            norm_mag = _NormalizerFactory.create('TB-Fam:MA-B')
        elif psname.sec == 'BO' and psname.dev == 'B':
            norm_mag = _NormalizerFactory.create('BO-Fam:MA-B')
        elif psname.sec == 'TS' and psname.dev == 'B':
            norm_mag = _NormalizerFactory.create('TS-Fam:MA-B')
        elif psname.sec == 'SI' and psname.dev == 'B1B2':
            norm_mag = _NormalizerFactory.create('SI-Fam:MA-B1B2')
        else:
            maname = psname.replace(':PS', ':MA').replace(':PU', ':PM')
            norm_mag = _NormalizerFactory.create(maname)
        return norm_mag

    @staticmethod
    def _get_devices(devname, proptype, auto_mon):
        # is dipole?
        if StrengthConv._get_dev_if_dipole(devname):
            return None, None

        # is insertion device?
        if StrengthConv._get_dev_if_id(devname):
            return None, None

        # is trim?
        status, dev_dip, dev_fam = \
            StrengthConv._get_dev_if_trim(devname, proptype, auto_mon)
        if status:
            return dev_dip, dev_fam

        # is booster ps?
        status, dev_dip = StrengthConv._get_dev_if_booster(
            devname, proptype, auto_mon)
        if status:
            return dev_dip, None

        # is others
        return StrengthConv._get_dev_others(devname, proptype, auto_mon), None

    @staticmethod
    def _get_dev_if_dipole(devname):
        if devname.dev in {'B', 'B1B2'}:
            # dipoles need no connectors
            return True
        return False

    @staticmethod
    def _get_dev_if_id(devname):
        if devname.dis == 'ID':
            # insertion devices need no connectors
            return True
        return False

    @staticmethod
    def _get_dev_if_trim(devname, proptype, auto_mon):
        if StrengthConv._is_trim(devname):
            # trims need dipole and family connectors
            dev_dip = PSProperty(
                'SI-Fam:PS-B1B2-1', 'Energy' + proptype, auto_mon)
            devname = devname.substitute(sub='Fam')
            dev_fam = PSProperty(devname, 'KL' + proptype, auto_mon)
            return True, dev_dip, dev_fam
        return False, None, None

    @staticmethod
    def _get_dev_if_booster(devname, proptype, auto_mon):
        if devname.startswith('BO'):
            if devname.dev == 'InjKckr':
                # BO injection kicker uses TB dipole normalizer
                dev_dip = PSProperty(
                    'TB-Fam:PS-B', 'Energy' + proptype, auto_mon)
            elif devname.dev == 'EjeKckr':
                # BO ejection kicker uses TS dipole normalizer
                dev_dip = PSProperty(
                    'TS-Fam:PS-B', 'Energy' + proptype, auto_mon)
            else:
                # other BO ps use BO dipoles as normalizer
                dev_dip = PSProperty(
                    'BO-Fam:PS-B-1', 'Energy' + proptype, auto_mon)
            return True, dev_dip
        return False, None

    @staticmethod
    def _get_dev_others(devname, proptype, auto_mon):
        if devname.startswith('LI'):
            return PSProperty('TB-Fam:PS-B', 'Energy' + proptype, auto_mon)
        if devname.startswith('TB'):
            # all TB ps other than dipoles need dipole connectors
            return PSProperty('TB-Fam:PS-B', 'Energy' + proptype, auto_mon)
        elif devname.startswith('TS'):
            # all TS ps use TS dipole
            return PSProperty('TS-Fam:PS-B', 'Energy' + proptype, auto_mon)
        elif devname.startswith('SI'):
            if devname.dev in {'InjDpKckr', 'InjNLKckr'}:
                # SI injection ps use TS dipole
                return PSProperty('TS-Fam:PS-B', 'Energy' + proptype, auto_mon)
            else:
                # other SI ps use SI dipole
                return PSProperty(
                    'SI-Fam:PS-B1B2-1', 'Energy' + proptype, auto_mon)
        return None

    @staticmethod
    def _is_trim(psname):
        check = (psname.sec == 'SI')
        check &= (psname.sub != 'Fam')
        check &= 'QS' not in psname.dev
        check &= 'Q' in psname.dev
        return check
